import java.lang.reflect.InaccessibleObjectException;
import java.net.InetSocketAddress;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.TimeUnit;

import com.github.benmanes.caffeine.cache.Cache;
import com.github.benmanes.caffeine.cache.Caffeine;
import com.ibm.darpc.DaRPCClientEndpoint;
import com.ibm.darpc.DaRPCClientGroup;
import com.ibm.darpc.DaRPCStream;

import org.apache.log4j.helpers.SyslogQuietWriter;

import hpdos.cache.SalCache;
import hpdos.handlers.NetworkHandler;
import hpdos.invalidationServer.InvalidationServer;
import hpdos.protocol.AckType;
import hpdos.protocol.Request;
import hpdos.protocol.RequestType;
import hpdos.protocol.Response;
import hpdos.protocol.RpcProtocol;

public class MetadataClient{
    private DaRPCStream<Request, Response> stream;
    private DaRPCClientEndpoint<Request, Response> clientEp;
    private  InvalidationServer invalidationServer;
    // SalCache cache;
    public Cache<String, String> cache;
    NetworkHandler networkHandler;

    public MetadataClient(Properties properties) throws Exception
    {
        String invalidationServerIP = String.valueOf((String)properties.get("app.HOST"));
        List<String> metadataMasters = new ArrayList<>();
        // String metadataMasterIp = String.valueOf((String)properties.get("app.MASTER_HOST1"));
        int cacheSize = Integer.valueOf((String)properties.get("app.CACHE_SIZE")); 
        int numberOfMasters = Integer.valueOf((String)properties.get("app.NO_OF_MASTERS"));
        this.networkHandler = new NetworkHandler(cacheSize, numberOfMasters);
        this.invalidationServer = new InvalidationServer(cache,invalidationServerIP, this.networkHandler);
        this.invalidationServer.acceptInvalidationConnections(numberOfMasters);
        for(int i=1;i<=numberOfMasters;i++){
            metadataMasters.add(String.valueOf((String)properties.get("app.MASTER_HOST"+i)));
        }
        for (String metadataMasterIp : metadataMasters) {
            this.invalidationServer.sendInvalidationRegistrationRequest(metadataMasterIp, invalidationServerIP);
            RpcProtocol rpcProtocol = new RpcProtocol();
            DaRPCClientGroup<Request, Response> group = DaRPCClientGroup.createClientGroup(rpcProtocol, 100, 0, 16, 16); 
            InetSocketAddress address = new InetSocketAddress(metadataMasterIp, 1920);
            clientEp = group.createEndpoint();
            clientEp.connect(address, 1000);
            stream = clientEp.createStream();
            NetworkHandler.streams.put(metadataMasterIp, stream);
            NetworkHandler.ips.add(metadataMasterIp);
            System.out.println("Connected");
        }
        this.cache = Caffeine.newBuilder()
                                            .expireAfterWrite(30, TimeUnit.SECONDS)
                                            .maximumSize(cacheSize)
                                            .build();
        
        
        
    }

    public byte[] get(byte[] key) throws Exception
    {
        System.out.println("Getting From local cache");
        // var value = this.cache.getIfPresent(key);
        String value = this.cache.getIfPresent(new String(key));
        if(value != null)
            return value.getBytes(StandardCharsets.UTF_8);
        System.out.println("Not found in cache Getting from server");  
        Request request= new Request();
        request.setRequestType(RequestType.GET);
        request.setKey(key);
        
        Response response = new Response();
        stream.request(request, response, false).get();
        if(response.getAck() == AckType.SUCCESS_WITH_VALUE){
            System.out.println("Adding to cache");
            this.cache.put(new String(key), new String(response.getValue()));
            return response.getValue();
        }

            
        return null;
    }

    public int put(byte[] key,byte[] value) throws Exception
    {
        System.out.println("Putting KV pair on server ....");
        Request request= new Request();
        request.setRequestType(RequestType.PUT);
        request.setKey(key);
        request.setValue(value);
        
        Response response = new Response();
        stream.request(request, response, false).get();
        if(response.getAck() == AckType.SUCCESS){
            this.cache.put(new String(key), new String(value));
            System.out.println("Updating cache..");
        }  
        return response.getAck();
    }

    public int delete(byte[] key) throws Exception
    {
        System.out.println("Invalidating Local Cache and sending delete request to server");
        // this.cache.invalidate(key);
        Request request= new Request();
        request.setRequestType(RequestType.DELETE);
        request.setKey(key);
        
        Response response = new Response();
        stream.request(request, response, false).get();
        return response.getAck();
    }
}