package hpdos.handlers;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;

import com.github.benmanes.caffeine.cache.Cache;
import com.github.benmanes.caffeine.cache.Caffeine;
import com.ibm.darpc.DaRPCClientEndpoint;
import com.ibm.darpc.DaRPCClientGroup;
import com.ibm.darpc.DaRPCFuture;
import com.ibm.darpc.DaRPCStream;
import com.ibm.darpc.DaRPCClientGroup.RpcClientFactory;

import org.apache.commons.math3.util.Pair;

import hpdos.cache.SalCache;
import hpdos.packet.Packet;
import hpdos.protocol.AckType;
import hpdos.protocol.Request;
import hpdos.protocol.RequestType;
import hpdos.protocol.Response;
import hpdos.protocol.RpcProtocol;


public class NetworkHandler {

    public static LinkedHashMap<String, DaRPCStream<Request, Response>> streams = new LinkedHashMap<String, DaRPCStream<Request, Response>>();
    public static LinkedList<String> ips =  new LinkedList<>();
    public int numberOfMasters;
    public Cache<String, String> cache;

    public NetworkHandler(int cacheSize, int numberOfMasters){
        this.numberOfMasters = numberOfMasters;
        this.cache = Caffeine.newBuilder().expireAfterWrite(30, TimeUnit.SECONDS).maximumSize(cacheSize).build();
    }
    
    // public boolean connectToServer() throws Exception{
    //     RpcProtocol rpcProtocol = new RpcProtocol();
    //     DaRPCClientGroup<Request, Response> group = DaRPCClientGroup.createClientGroup(rpcProtocol, 100, 0, 16, 16); 
    //     InetSocketAddress address = new InetSocketAddress("192.168.200.50", 1919);
    //     DaRPCClientEndpoint<Request, Response> clientEp = group.createEndpoint();
    //     clientEp.connect(address, 1000);
    //     DaRPCStream<Request, Response> stream = clientEp.createStream();
        
    //     if(!stream.isEmpty()){
    //         this.streams.put("192.168.200.50", stream);
    //         return true;
    //     }
    //     else{
    //         return false;
    //     }
    // }

    // public Response create(Packet packet) throws IOException, InterruptedException, ExecutionException{
    //     SalCache.cache.put(packet.getKey(), packet.getValue());
    //     Request request = new Request();
    //     Response response = new Response();
    //     request.setRequestType(RequestType.GET);
    //     request.setKey(packet.getKey());
    //     request.setValue(packet.getValue());
    //     int partition = partition(packet.getKey());
    //     response = streams.get(ip_addresses.get(partition)).request(request, response, false).get();
    //     return response;
    // }

    public byte[] get(byte[] key) throws Exception
    {
        System.out.println("Getting From local cache");
        // var value = this.cache.getIfPresent(key);
        String value = this.cache.getIfPresent(new String(key));
        if(value != null)
            return value.getBytes(StandardCharsets.UTF_8);
        System.out.println("Not found in cache, Getting from server");  
        Request request= new Request();
        request.setRequestType(RequestType.GET);
        request.setKey(key);
        
        Response response = new Response();


        // Partitioning
        int masterPartition = partition(key);
        String ip = NetworkHandler.ips.get(masterPartition);

        // Load balancing - TODO

        DaRPCStream<Request, Response> stream = NetworkHandler.streams.get(ip);
        stream.request(request, response, false).get();

        if(response.getAck() == AckType.SUCCESS_WITH_VALUE){
            System.out.println("Adding to cache");
            this.cache.put(new String(key), new String(response.getValue()));
            return response.getValue();
        }

            
        return null;
    }

    public int put(byte[] key,byte[] value) throws Exception
    {
        System.out.println("Putting KV pair on server ....");
        Request request= new Request();
        request.setRequestType(RequestType.PUT);
        request.setKey(key);
        request.setValue(value);
        
        Response response = new Response();
       // Partitioning
       int masterPartition = partition(key);
       String ip = NetworkHandler.ips.get(masterPartition);

       // Load balancing - TODO

       DaRPCStream<Request, Response> stream = NetworkHandler.streams.get(ip);
       stream.request(request, response, false).get();

        if(response.getAck() == AckType.SUCCESS){
            this.cache.put(new String(key), new String(value));
            System.out.println("Updating cache..");
        }  
        return response.getAck();
    }

    public int delete(byte[] key) throws Exception
    {
        System.out.println("Invalidating Local Cache and sending delete request to server");
        // this.cache.invalidate(key);
        Request request= new Request();
        request.setRequestType(RequestType.DELETE);
        request.setKey(key);
        
        Response response = new Response();
        // Partitioning
        int masterPartition = partition(key);
        String ip = NetworkHandler.ips.get(masterPartition);

        // Load balancing - TODO

        DaRPCStream<Request, Response> stream = NetworkHandler.streams.get(ip);
        stream.request(request, response, false).get();
        return response.getAck();
    }
    // public Pair<byte[], byte[]> read(byte[] key) throws InterruptedException, ExecutionException, IOException{
    //     Pair<byte[], byte[]> kPair;
    //     byte[] value = SalCache.cache.getIfPresent(key);
    //     if (value == null) {
    //         // Add code to read from NIC/SSD
    //         Request request = new Request();
    //         Response response = new Response();
    //         request.setRequestType(RequestType.GET);
    //         request.setKey(key);
    //         request.setValue(null);
    //         int partition = partition(key);
    //         response = streams.get(ip_addresses.get(partition)).request(request, response, false).get();
    //         value = response.getValue();
    //         kPair = new Pair<byte[],byte[]>(key, value);
    //         return kPair;
    //     }
    //     else {
    //         // Add code to return key,value from sal cache
    //         kPair = new Pair<byte[],byte[]>(key, value);
    //         return kPair;
    //     }
        
    // }

    // public Response update(Packet packet) throws InterruptedException, ExecutionException, IOException{
    //     SalCache.cache.put(packet.getKey(), packet.getValue());
    //     Request request = new Request();
    //     Response response = new Response();
    //     request.setRequestType(RequestType.PUT);
    //     request.setKey(packet.getKey());
    //     request.setValue(packet.getValue());
    //     int partition = partition(packet.getKey());
    //     response = streams.get(ip_addresses.get(partition)).request(request, response, false).get();
    //     return response;
    // }

    // public Response delete(Packet packet) throws InterruptedException, ExecutionException, IOException{
    //     SalCache.cache.put(packet.getKey(), packet.getValue());
    //     Request request = new Request();
    //     Response response = new Response();
    //     request.setRequestType(RequestType.DELETE);
    //     request.setKey(packet.getKey());
    //     request.setValue(null);
    //     int partition = partition(packet.getKey());
    //     response = streams.get(ip_addresses.get(partition)).request(request, response, false).get();
    //     return response;
    // }

    public int partition(byte[] key)
    {
        int hash = 7;
        hash = 31 * hash + (key == null ? 0 : key.hashCode());
        return hash % numberOfMasters;
    }

}
