package hpdos.invalidationServer;

import java.io.IOException;
import java.io.ObjectOutputStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;

import com.github.benmanes.caffeine.cache.Cache;
import com.ibm.darpc.DaRPCServerEndpoint;
import com.ibm.darpc.DaRPCServerGroup;
import com.ibm.disni.RdmaServerEndpoint;

import hpdos.handlers.NetworkHandler;
import hpdos.protocol.InvalidationRequest;
import hpdos.protocol.InvalidationResponse;
import hpdos.services.InvalidationService;

public class InvalidationServer
{
    Cache<String, String> cache;
    RdmaServerEndpoint<DaRPCServerEndpoint<InvalidationRequest, InvalidationResponse>> serverEp;
    public InvalidationServer(Cache<String, String> cache2,String invalidationServerIP, NetworkHandler networkHandler)throws Exception
    {
            this.cache = cache2; 
            long[] clusterAffinities =  new long[1];
            clusterAffinities[0] = 1<<1;
            DaRPCServerGroup<InvalidationRequest, InvalidationResponse> group = null;
            InvalidationService  service = new InvalidationService(networkHandler);

            group = DaRPCServerGroup.createServerGroup(service, clusterAffinities, -1, 1, true, 2, 2, 2, 1);    
            serverEp = group.createServerEndpoint();
            InetSocketAddress address = new InetSocketAddress(invalidationServerIP, 1922);
            serverEp.bind(address, 2);
    }
    public void acceptInvalidationConnections(int numOfConnections) throws Exception
    {
        Runnable runnable = ()->
            {
                
                try{
                    int conns = numOfConnections;
                    System.out.println("Waiting for server to send connection request for invalidation");
                    while (conns>0) {
                        serverEp.accept();
                        conns -- ;      
                    }
                    
                    System.out.println("Got Connected for Invalidation");
                }catch(Exception e){
                    e.printStackTrace();
                }
            };
        var t= new Thread(runnable);
        t.setName("Server Connection");
        t.start();
        System.out.println("started Invalidation server");
       
    }

    public void sendInvalidationRegistrationRequest(String masterIpAddress, String hostIpAddress) throws IOException{
        System.out.println("Sending invalidation registration request.");
        Socket socket = new Socket(masterIpAddress, 9875);
        ObjectOutputStream oos = null;
        oos = new ObjectOutputStream(socket.getOutputStream());
        oos.writeObject(hostIpAddress);
        socket.close();
    }
    
}
